/******************************************************************************
* $Id: scoring.sas,v 1.1.2.9.4.1 2012/03/07 13:36:13 saswai Exp $
*
* Copyright(c) 2011 SAS Institute Inc., Cary, NC, USA. All Rights Reserved.
*
* Name: scoring.sas
*
* Purpose: Score claims based on expert model
*
* Author: Steve Uhorchak (stuhor)
*
* Support: SAS(r) Advanced Analytics Lab
*
* Input:	sp_prep.sas7bdat -- the second pass prep table
*			threshold -- number of risk factors w/o associated alerts needed to create a new alert
*        
*
* Output: stage.enriched_alert_summary.sas7bdat -- this table has been updated to include the
*									 	  scores and ranks for each actionable_entity
*									 	  within each entity_nm
*		  stage.alert_entity.sas7bdat --  this table contains the data rolled into each 
*										  unique actionable_entity, along with the corresponding
*										  scores and rank
*
* Parameters: 
*
* Dependencies/Assumptions: 
*
* Usage: The input table can be an alert_summary table from the current run only, or aggregated alert summary 
*		 tables over a designated period of time
*
* History:
* ddmmmyyyy userid  description (Change Code)
* 
\******************************************************************************/

%macro scoring(input_lib=STAGE, input_table=ENRICHED_ALERT_SUMMARY, output_lib=STAGE, threshold=0);
	%let macroname=%get_macroname;
	%put NOTE:(FCM) Entering &macroname;

/* These are locally created variables that should remain in the program
   Please do not comment out
*/

%let wt_severity_score = 60 ;   /* default=60 */
%let wt_n_alerts_score = 20 ;   /* default=20 */
%let wt_amt_loss_score = 20 ;   /* default=20 */

/* Drop unnecessary variables from the input table */

data work.summary; 
	set &output_lib..SP_PREP (keep= entity_sk scenario_sk entity_variable_sk scenario_weight_no scenario_loss_amt scenario_severity_score_no
									segment_id record_version_no entity_nm);
run;

/* Create a table for every ENTITY_SK */

%do i=1 %to &entity_nobs;

	proc sql noprint;
		create table entity_type&i as
		select summary.*
		from summary
		where entity_sk=&&entity_sk&i
		order by scenario_sk, entity_sk;
	quit;
%end;

%do i=1 %to &entity_nobs;

	/* Normalize severity scores */

	proc summary data=entity_type&i noprint nway;
		class scenario_sk;
		var scenario_severity_score_no scenario_weight_no;
		output out=scen_norm&i (drop=_type_ _freq_)	max=scen_sev_score_no_max_grp 
													median= scen_sev_score_med scen_weight_no_med; 
	run;

	data entity_type&i;
		merge entity_type&i scen_norm&i;
		by scenario_sk;

		if scen_sev_score_no_max_grp=. or scen_sev_score_no_max_grp=0 
			then put "FCM_NOTE: All scenario severity scores are zero or missing.";

		else do;
			if scenario_weight_no=. or scenario_weight_no=0 
				then scenario_weight_no = scen_weight_no_med;

			if scen_weight_no_med=.  or scen_weight_no_med=0 
				then norm_sev_score_no= abs(scenario_severity_score_no/scen_sev_score_no_max_grp);

   			else norm_sev_score_no= abs(scenario_severity_score_no/scen_sev_score_no_max_grp) * scenario_weight_no;
		end;
	run;

	/* Table claim&i is scenario_sk, entity_sk
	   Calculate the following with e at the entity level (_grp):
			Scenario_nvio[s,e] = Number of times scenario s was violated by entity e.
			Scenario_severity[s,e] = Severity of the violations of scenario s by entity e. 
			Scenario_amtloss[s,e] = Maximum potential amount lost associated with the violations of scenario s by entity e.
	*/
	
	proc summary data=entity_type&i noprint nway;
		id scenario_weight_no;
		class scenario_sk entity_sk;
		var norm_sev_score_no scenario_loss_amt;
		output out=claim&i (drop=_type_ rename=(_freq_=scenario_nvio_grp))	std=scen_sev_score_no_stddev_grp scen_loss_amt_stddev_grp
																			max=scen_sev_score_no_max_grp scen_loss_amt_max_grp
																			p95=scen_sev_score_no_p95_grp scen_loss_amt_p95_grp
																			sum=scen_severity_score_no_grp_total scenario_loss_amt_grp_total;
	run;

	/* Table nvio&i is scenario_sk, entity_variable_sk 
	   Calculate the following with e at the entity variable level (_i):
			Scenario_nvio[s,e] = Number of times scenario s was violated by entity variable e.
			Scenario_severity[s,e] = Severity of the violations of scenario s by entity variable e. 
			Scenario_amtloss[s,e] = Maximum potential amount lost associated with the violations of scenario s by entity variable e.
	*/

	proc summary data=entity_type&i noprint nway;
		id scenario_weight_no scenario_loss_amt segment_id record_version_no entity_nm;
		class scenario_sk entity_variable_sk;
		var norm_sev_score_no scenario_loss_amt;
		output out=nvio&i (drop=_type_ rename=(_freq_=scenario_nvio_i))	std=scen_sev_score_no_stddev_i scen_loss_amt_stddev_i
																		max=scen_sev_score_no_max_i scen_loss_amt_max_i 
																		p95=scen_sev_score_no_p95_i scen_loss_amt_p95_i 
																		sum=scen_sev_score_no_sum_i scen_loss_amt_sum_i;
	run ;

	/* Merge the data sets so that relevant variables from claim&i and nvio&i are on the same dataset
	   Calculate:
			Scenario_nvio score for each [s,e]
			Scenario_severity score for each [s,e] 
			Scenario_amtloss for each [s,e]
			Actionable_entity score = Scenario_nvio+Scenario_severity+Scenario_amtloss for each [s,e]
	*/

	data _alert_entity&i(rename=(scenario_nvio_i=total_n_alerts_no 
							     scen_loss_amt_sum_i=scenario_total_loss_amt )  
						         );
		keep entity_variable_sk 
             scenario_nvio_i
			 scen_loss_amt_sum_i
			 entity_sk
			 entity_nm
			 record_version_no
			 segment_id
			 severity_score_no 
			 n_alerts_score_no
			 amt_loss_score_no
			 actionable_entity_score_no
		;
		merge claim&i nvio&i;
		by scenario_sk;
		
		n_alerts_score_no = round((&wt_n_alerts_score * min(1,(scenario_nvio_i / scenario_nvio_grp))),.01);
		
		if scenario_loss_amt_grp_total=. or scenario_loss_amt_grp_total=0 
			then amt_loss_score_no=0;
		else amt_loss_score_no = round(&wt_amt_loss_score * min(1,(scen_loss_amt_sum_i / scenario_loss_amt_grp_total) ),.01);
		
		if scen_severity_score_no_grp_total=. or scen_severity_score_no_grp_total=0
			then severity_score_no=0;
		else severity_score_no = round(&wt_severity_score * min(1,(scen_sev_score_no_sum_i / scen_severity_score_no_grp_total ) ),.01);
		
		actionable_entity_score_no = severity_score_no + n_alerts_score_no + amt_loss_score_no ;
		
	run; 



	/* Sort entities by actionable_entity_score */
	proc sort data=_alert_entity&i;
		by descending actionable_entity_score_no descending scenario_total_loss_amt ;
	run ;

	/* Creates a rank by actionable entity score, with the highest score having a rank of 1 */
	data _alert_entity&i;
		set _alert_entity&i;
		by descending actionable_entity_score_no descending scenario_total_loss_amt ;   
		retain actionable_entity_rank_no ;
		if first.actionable_entity_score_no then actionable_entity_rank_no = _n_ ;
	run ;
	
	/* Append all _alert_entity&i tables to create an aggregated table of all actionable entities */
	%if &i > 1 %then %do;
		proc append base=_alert_entity1 data=_alert_entity&i;
		run;
    %end;
%end;

/* Obtain total score for all entities */
data _null_;
	set _alert_entity1 end=last;
	total_act_ent_score_no + actionable_entity_score_no;
	if last then call symputx('total_act_ent_score_no', total_act_ent_score_no);
run;

/* Determine overall score for each individual entity */
proc sort data= _alert_entity1;
	by entity_sk entity_variable_sk;
run;

data _alert_entity1;
	set _alert_entity1;
	by entity_sk entity_variable_sk;
	if first.entity_variable_sk then
		ent_var_total_score = 0;
	ent_var_total_score+actionable_entity_score_no;
	if last.entity_variable_sk then do;
		overall_actionable_entity_score = ent_var_total_score/&total_act_ent_score_no;
		output;
	end;
run;
	
/* Calculate an overall ranking across all actionable entities */
proc sort data=_alert_entity1;
	by descending overall_actionable_entity_score descending scenario_total_loss_amt; 
run;

/* Creates an overall rank regardless of entity_nm, with the highest score having a ranking of 1 */
data _alert_entity1;
	set _alert_entity1;
	by descending overall_actionable_entity_score descending scenario_total_loss_amt; 
	retain overall_actionable_entity_rank;
	if first.overall_actionable_entity_score then overall_actionable_entity_rank =_n_;
run;

/* Write the _alert_entity aggregated table to the stage directory */
data &output_lib.._alert_entity;
	set _alert_entity1;
run;

	%put NOTE:(FCM) Exiting &macroname;
%mend scoring;
